/* fillform - Part of Form Fill-r program
*/
#include "globals.h"

extern void give_error(char *text);
extern void display_status();
extern void recalc();
extern void showreal(int row, int col, struct field *afield);
extern void get_value_from_field(struct field *afield);
extern void center (int line, char *text);
extern void fkeydefn (int line, int col, int fkey, char *text);

/* in saveform.c: */
extern char Unable_To_Open[];

int cur_col;        /* First column of current field */
struct field *ref_field;
int top_line;       /* Line at top of window. */
int win_size;       /* Number of lines in window */
int scrn_bot;       /* Screen line which is bottom of window */
int scrn_top;       /* Screen line which is top of window */


/* REF_TOP ------------------------------

	Moves the reference point to the top of the form.
*/
void ref_top()
{
	ref_field = first_field;
	while (ref_field -> entry_row < 0)	/* Ignore Constant data lines */
		ref_field = ref_field -> next;
}

/*-------------------------- go_left  ---------------------------*/
void go_left()
/*  Move to prior data field, if none, do not move. */
{   struct field *p;

	p = cur_field;
	while ((p = p -> prev) != NULL)
	  if ((p -> type > 0) && (p -> prot ==0))	/* only move into */
	     {	cur_field = ref_field = p;		/* fields which are */
		return; }				/* non-protected */
}

/*-------------------------- go_right ---------------------------*/
void go_right()
/*  Move to next field, if none, do not move. */
{   struct field *p;

	p = cur_field;
	while ((p = p -> next) != NULL)
	  if ((p -> type > 0) && (p -> prot ==0))	/* only move into */
	     {	cur_field = ref_field = p;		/* fields which are */
		return; }				/* non-protected */
}

/*-------------------------- go_up  ---------------------------*/
void go_up(int nl)
{
    struct field *p, *prev_field;
    int target_line, target_col, found_field = 0;

	/* Find field on target line, to left of (or at) target column */
	target_line =  (cur_field ->entry_row - nl > 0) ? 
			cur_field ->entry_row - nl : 0;
	if (target_line < 0) target_line = 0;

	target_col = cur_field -> entry_col;

	prev_field = p = cur_field;	/* Assume no previous field */

	while ((p = p -> prev) != NULL )	/* Stop at top of list */
	  { if (p -> prot ==0 && p -> type > 0)
			/* Found some data field on or above current line */
	      {	if (found_field > 0 && p -> entry_row != target_line)
			break;	/* We previously found a field on the right
				   line; this one is above that line, so use
				   what we found previously. */
		prev_field = p;
		if (p -> entry_row <= target_line && found_field++ == 0)
			target_line = p -> entry_row;
	        if (p -> entry_row == target_line &&
		    p -> entry_col <= target_col)
			break;		/* Found the right one */
	      }
	  }

	cur_field = ref_field = p = prev_field;

	/* Set ref_field to a data field on the line previous to
	   the new current field */
	while ( (p=ref_field -> prev) != NULL &&
		(p -> entry_row == cur_field -> entry_row || p -> type <= 0))
	   ref_field = p;
	if (ref_field == cur_field)	/* No fields above this one, display
					   to the top of the file */
	  ref_top();
}

/*-------------------------- go_down  ---------------------------*/

void go_down(nl)
int nl;
{
    struct field *p, *next_field;
    int target_line, target_col, found_field=0;

	/* Find field on target line, to left of (or at) target column */
	target_line = cur_field -> entry_row + nl;
	target_col  = cur_field -> entry_col;

	if (target_line > last_line) target_line = last_line;

	next_field = p = cur_field;	/* Assume no field below current */

	while ((p = p -> next) != NULL )	/* Stop at end of list */
	  { if (p -> prot ==0 && p -> type > 0)
			/* Found some data field on or below current line */
	      {	if (found_field > 0 && p -> entry_row != target_line)
			break;	/* We previously found a field on the right
				   line; this one is below that line, so use
				   what we found previously. */
		next_field = p;
		if (p -> entry_row >= target_line && found_field++ == 0)
			target_line = p -> entry_row;
	        if (p -> entry_row == target_line &&
		    p -> entry_col + strlen(p -> mask) > target_col)
			break;		/* Found the right one */
	      }
	  }

	cur_field = ref_field = p = next_field;

	/* Set ref_field to the next field below this one,
	   so the window will scroll down enough to see below any
	   blank lines */
	while ( (p=ref_field -> next) != NULL &&
		(p -> entry_row == cur_field -> entry_row || p -> type <= 0))
	   ref_field = p;

	if (ref_field == cur_field)	/* No field below this, go to bottom of file */
	 ref_field = last_field;
}

/*---------------------------- show_window ---------------------------*/
void show_window(unconditional) /* called only from input_scan */
int unconditional; /* show window unconditionally */

/* rows refer to the screen display 
   lines refer to the user defined data lines
   fields refer to the user defined data fields
*/
{
int new_line;       /* Temp variable */
int bot_line;
int old_top,old_bot,start_fill,end_fill;
int field_line;
int bkg_start;		/* start index in background string */
char *pbg, *pfg;	/* background & foreground scanners */
char temp[140];		/* for printing strings */
struct field *background, *foreground;
    
    if (get_mode != 1) return;
    old_top=cur_top;
    old_bot=cur_top + win_size - 1;
    new_line = ref_field -> entry_row;
    if (new_line < 0) new_line = 0;

    if (unconditional)
    {   start_fill = old_top;
        end_fill   = old_bot;
        top_line   = new_line;
        scroll(scrn_top,scrn_bot,ERASE,S_UP); /* erase all */
    }
    else
    {   /* Test for new line below window. */
        if (new_line > old_bot)
        {   bot_line = new_line;
            top_line = new_line - win_size + 1;
            if (top_line <= old_bot)
                scroll(scrn_top,scrn_bot,top_line-old_top,S_UP);
            else
                scroll(scrn_top,scrn_bot,ERASE,S_UP); /* erase all */
            end_fill   = new_line;
            /* Do not allow overscan of window */
            start_fill = (old_bot+1) > top_line ? old_bot+1: top_line;
        }
        /* Test for new line above the window. */
        else if (new_line < old_top)
        {   top_line = new_line;
            bot_line = new_line + win_size - 1; 
            if (bot_line >= old_top)
                scroll(scrn_top,scrn_bot,old_top-top_line,S_DOWN);
            else
                scroll(scrn_top,scrn_bot,ERASE,S_UP); /* erase all */
	    start_fill = top_line;
            /* Do not allow overscan of window */
            end_fill   = (old_top-1) < bot_line ? old_top-1: bot_line;
        }
        /* Already in the window */
        else return;
    }

    /* Put the new lines on the screen */
    cur_top=top_line;     /* Establish new top line */
    /* find the top field on the current line. */

    background = first_field;
    while (background != NULL && background -> entry_row < start_fill)
	background = background -> next;
    /* background now points to the first field on the line, which
       should contain the background information */

    if (background == NULL) return;	/* No fields to display */

    while (background != NULL &&
	   (field_line = background -> entry_row) >= start_fill &&
	   field_line <= end_fill)
      {	setcur(field_line+scrn_top-top_line,0);
	foreground = background -> next;	/* First actual field on line, if any */
	pbg = background -> entry_line;		/* Pointer to background */
	bkg_start=0;
	strcpy(temp,pbg);		/* Make a copy of background */

	while (foreground != NULL && 	/* while still foreground (fields) */
	       foreground -> entry_row == background -> entry_row)
					/* and they're on this line */
	  { if (foreground -> type > 0)	/* For each data field: */
	      {	temp[foreground -> entry_col - bkg_start]=0;
					/* don't print bkg over fg */
		wstr(temp,1);

		pfg = foreground -> work;
		bkg_start += strlen(pfg) + strlen(temp);
					/* new background begins past
					   previous bkg and this field */
		
		if (!(foreground -> prot))	/* Don't highlight protected fields */
			wchar(1);		/* Inverse video for field */
		wstr(pfg,1);
		if (!(foreground -> prot))	/* Don't highlight protected fields */
			wchar(2);

		temp[0]=0;		/* Assume no background */
					/* if any background left, copy it */
		if (bkg_start <= strlen(pbg))
		  strcpy(temp,&(pbg[bkg_start]));

	      }
	    foreground = foreground -> next;
	  }
	wstr(temp,1);			/* Write out remainder of background */
	background = foreground;	/* Go beyond last field on current line,
					   presumably background of next line. */
      }
}

/*---------------------------- fillform --------------------------*/

int fillform()

{
int     help_line, x, ok_to_print,linesmoved;
#ifdef DEMO
static int print_count=0;		/* Print only one form in Demo version */
#endif

	cur_top = 0;
	scrn_top=win_sr;
	scrn_bot=win_lr;
	win_size = win_nr;

	/* clear screen and show initial buffer information. */
	cls();
	center(0,form_title);

	help_line = scrn_bot+1;
	fkeydefn(help_line, 0, 2, "Print");
	fkeydefn(help_line, 9, 3, "Copy Prior Field");
	fkeydefn(help_line,29, 5, "Erase All Fields");
	help_line++;
	fkeydefn(help_line, 9, 4, "Copy Prior Form");
	fkeydefn(help_line,29, 9, "Recalculate");
	fkeydefn(help_line,48,10, MainMenu);

    /* Setup to show initial window on form file. */
    cur_field = first_data_field;
    ref_top();
    show_window(1);

/*  This loop gathers the user input.
    The loop index is cur_field which is the current data entry field. 
*/
    for ( ; ; ) /* Exit loop via F10 or EOF */
    {
        cur_col = cur_field->entry_col;
        show_window(0);
	display_status();
        wchar('\01');		/* Inverse video attribute while editing*/
        x = select_f(cur_field);
        wchar('\02');		/* Normal video attribute */
	stat_line = blanks;	/* Reset status line to blanks */
        switch (x)
        {
        /* Copy current field */
        case F3:
	    strcpy(cur_field -> work, cur_field -> last);
	    get_value_from_field(cur_field);
            break;
        /* Copy all fields */
        case F4:
	    for (cur_field = first_data_field; cur_field != NULL; cur_field = cur_field -> next)
	      {	if (cur_field -> work != NULL)	/* for each data field */
		  { strcpy(cur_field -> work, cur_field -> last);
		    get_value_from_field(cur_field);
		  }
						/* copy last value */
	      }
            cur_field = first_data_field;
	    ref_top();
            cur_top = 0;
            show_window(1);
            break;
        /* Print form and blank fields */
        case F2:
            if (form_file_open)
            {struct field *afield;
		recalc();
		afield = first_data_field;
		ok_to_print = 1;	/* Assume all protected fields filled */
		while (afield != NULL)
		  { if (afield -> reqd && strcmp(afield -> work, afield -> mask)==0)
		      {	give_error("\7Required Field Not Filled In");
			linesmoved = afield -> entry_row - cur_field -> entry_row;
			/* Move up/down to redetermine ref_field... */
			if (linesmoved > 0)	/* moved down */
				go_down(linesmoved);
			if (linesmoved < 0)	/* moved  up */
				go_up(-linesmoved);
			cur_field = afield;	/* then go to specific field */
			ok_to_print = 0;
			break;
		      }
		    afield = afield -> next;
		  }
		if (!ok_to_print)
		   break;
#ifdef DEMO
		/* Skip printing the form if we've already printed one */
		if (!print_count++)
#endif
		if (!print_form())	/* If successful printing form... */
			give_error("Printed");
		else
			give_error(Unable_To_Open);

		/* Copy all data fields */
		for (cur_field = first_data_field; cur_field != NULL; cur_field = cur_field -> next)
		   {	if (cur_field -> work != NULL)	/* for each data field */
			strcpy(cur_field -> last, cur_field -> work);
							/* save current value */
		   }
            }  /* NOTE!!! Falls through to F5. */
        /* Blank out all data fields */
        case F5:
	    for (cur_field = first_data_field; cur_field != NULL; cur_field = cur_field -> next)
	      {	if (cur_field -> work != NULL)	/* for each data field */
		  { strcpy(cur_field -> work, cur_field -> mask);
						/* erase & set to mask */
		    cur_field -> value = 0;
		    showreal(-1,-1,cur_field);
		  }
	      }
            cur_field = first_data_field;
	    ref_top();
            cur_top = 0;
            show_window(1);
            break;
	case F9:
	    recalc();
	    break;
        /* Select prior field */
	case SH_TAB:
        case CTL_LEFT:
            go_left();
            break;
        /* Select field above current */
        case UP:
            go_up(1);
            break;
        /* Scroll up 19 lines */
        case PGUP:
            go_up(win_nr-2);
            break;
        default:
            if ( x > 127 ) break;
        /* Select next field */
        case CTL_RIGHT:
        case TAB:
            go_right();
            break;
        /* Select field below current */
        case DOWN:
            go_down(1);
            break;
        /* Scroll down 19 lines */
        case PGDN:
            go_down(win_nr-2);
            break;
	/* Exit form filling operation */
        case F10:
        case EOF:
            return(x);
        }
    }
/* Return via F10 or EOF */
}
