/*

	PROCFDF.C

	Process the form definition -- create fields, and load
	fields, constants, and background lines into them.

	Part of Form Filler Program
*/

#include "globals.h"
#include <ctype.h>

/* #define DEBUG */

extern fatal_line(char *the_bad_line,char *message,int err_no);
extern int keyword(char *keywords[], char **input);
extern int punct(char c);

static char no_posn[] = {"No position given & "};
static char no_field[] = {"No data field & @"};
static char no_cnst[] =  {"No constant data & @"};
static int rw = 0;
static int cl = 0;
static char *pl;

/* NEWVAL -------------------------------------------

	Gets absolute or relative line number

*/
int newval(int nr)
{   int tmp, typ;
    ++pl;
    typ = tmp = 0;
    while (isspace(*pl)) ++pl; /* Skip over spaces */
    if (*pl == '+') {++pl; ++typ;} /* Positive relative */
    else
    if (*pl == '-') {++pl; --typ;} /* Negative relative */
    while (isdigit(*pl))
           tmp = tmp * 10 + (*pl++ & 0xf);
    if(typ) return(nr + tmp * typ);	/* Relative number */
    else return(tmp - 1);		/* Absolute number */
}

#ifdef DEBUG
/* SHOW_FIELD_LIST ----------------------------------


	Prints out the field lists.
*/
show_field_list()
{   struct field *afield;

	printf("\nPress any key to begin Field List...\n");
	gtkey();

	afield = first_field;
	while (afield != NULL)
	    {	printf("(%3d,%3d) -> (%3d,%3d), Type %1d",
			afield -> entry_row, afield -> entry_col,
			afield -> print_row,afield->print_col, afield -> type);
		if (afield -> mask)
		  printf("%s\n",afield -> mask);
		else
		  printf("%s\n",afield -> entry_line);
		afield = afield -> next;
	    }

	printf("\nPress any key to begin Print List...\n");
	gtkey();

	afield = first_print_field;
	while (afield != NULL)
	    {	printf("(%3d,%3d) -> (%3d,%3d), Type %1d ",
			afield -> entry_row, afield -> entry_col,
			afield -> print_row,afield->print_col, afield -> type);
		printf("%s\n",afield -> mask);
		afield = afield -> next_print;
	    }
}
#endif

/* NEW_FIELD ----------------------------------------

	Returns the pointer to a newly-created field.

*/

static struct field *new_field(int the_row, int the_col)
{
    struct field *p;

	p = calloc(1,sizeof(struct field));
	if (last_field == NULL)		/* no fields defined yet */
	    {	first_field = last_field = p;
		p -> next = p -> prev = NULL;
	    }
	else
	    { if (the_row >= 0)	/* Put everything but constants @ end of list */
		{ last_field -> next = p;
		  p -> prev = last_field;
		  last_field = p;
		}
	      else		/* Put constants together @ head of list */
	        { p -> next = first_field;	/* Insert before first */
		  first_field -> prev = p;
		  p -> prev = NULL;
		  first_field = p;
		}
	    }
	p -> work = p -> mask = p -> last = p -> entry_line =
		    p -> formula = NULL;
	p -> next_print = NULL;	
	*(p -> name) = '\0';

	/* Assume that fields print in their entry positions (SCREEN mode) */
	p -> entry_row = p -> print_row = the_row;
	p -> entry_col = p -> print_col = the_col;

	p -> type = p -> prot = p -> reqd = 
		p -> ucaseonly = p -> rightjust = p -> non_printing = 0;
	p -> value = 0;

	return(p);
}

/* INSERT_IN_PRINT_LIST ------------------------------

	Inserts the given field in the print list
*/

insert_in_print_list(struct field *new_field)
{
    struct field *p, *q;	/* q trails p by one node */

	p = first_print_field;
	q = NULL;	/* no previous fields */

	if (p == NULL)	/* Nothing yet in print list */
	    {	first_print_field = new_field;
		return;
	    }

	while (1)
	    {	if (p == NULL ||	/* New node goes at end of list */
		    (new_field -> print_row < p -> print_row) ||
					/* Is on line above this node */
		    ((new_field -> print_row == p -> print_row) &&
		     (new_field -> print_col < p -> print_col)))
					/* Is on same line, but to left
						of this node */
		  /* Find the node that the new node precedes. */
		  { new_field -> next_print = p; /* Even works if p is NULL */
		    if (q == NULL)	/* goes before current 1st entry */
			first_print_field = new_field;
		    else
			q -> next_print = new_field;

		    /* Put new field between nodes p and q */
		    return;
		  }
		q = p; p = p -> next_print;
	    }
}

/* MAKECOPY --------------------------------

	Makes a copy of the field in an allocated memory space

*/
char *makecopy(char *field_text,int fl)
{
    char *place;

	if ((place = malloc(fl+1)) == NULL)
	    fatal(err_mem,2);
	memcpy(place,field_text,fl);
	*(place+fl)='\0';
	return(place);
}

/* DIDNT_POSN -------------------------------

	Error -- user didn't specify where to position the given
	field.

*/
static didnt_posn(struct field *oops)
{
    char the_error[80], the_line[80];

	strcpy(the_error,no_posn);
	if (oops -> work == NULL)	/* If a constant data line */
	    {	strcat(the_error,"constant");
		strcpy(the_line,"@");
	    }
	else
	    {	sprintf(the_error+strlen(the_error),"field, on screen line %d",
			oops -> entry_row+1);
		the_line[0]='\0';
		if (screen_mode || data_mode) return;
		/* Don't need to position fields in screen or data modes */
	    }
	strcat(the_line,oops -> mask);
	fatal_line(the_line,the_error,23);

}

/* CHECK_POSN ------------------------------

	Checks the position which the user has specified

*/
static check_posn(char *lp)
{
    if (rw < 0 || rw >= MAX_ROWS) fatal_line(lp,"~Row #",24);
    if (cl < 0 || cl >= MAX_COLS) fatal_line(lp,"~Column #",24);
}

/* VALID_DOLLARS ---------------------------

	Tells us whether a mask is a valid dollars field.

*/
static int valid_dollars(char *mask)
{  char *p;
   int digits=0, decpt=0;

	p = mask;
	while (*p)
	  switch (*p++)
	    { case '_':		/* Allow underlines */
		if (decpt)	/* Count up the digits after the decimal point */
		  digits++;
	      case ',':		/* Allow commas (?) */
		break;
	      case '.':
		if (decpt++)
		  return(0);	/* Already had a decimal point! */
		break;
	      default:
		if (p != mask+1)/* Allow other characters only as initial */
		  return(0);	/* char, & not in middle of field */
	    }
	return(digits == 0 || digits == 2);
	/* Allow field to be floating (_____) or have 2 places (____.__) */
}

/* VALID_DATE_FIELD ---------------------------

	Tells us whether a mask is a valid date field.

*/
static int valid_date_field(char *mask)
{  char *p;
   int digits=0;

	p = mask;
	while (*p != '\0')
		digits += *p++ == '_';	/* count underlines */

	return(digits == 6);		/* Must be exactly 6 spaces */
}

static int valid_integer_field(char *mask)
{
	/* Valid if no dots or dashes */
	return (strchr(mask,'.')==NULL &&
		strchr(mask,'-')==NULL);
}

/* SETSCRN ----------------------------------

	Process one input entry screen line.
	Create fields for each field in the line.

*/

int setscrn(lp)
char              *lp;

{
int     field_count = 0;
int	ln, fl, row, x, start;
struct field *afield;
char	*eline;

    if (cur_field != NULL)
	didnt_posn(cur_field);

	/* Find entry-screen row */
    row = (last_field == NULL)?0:last_field -> entry_row+1;
    last_line = row;		/* Determine the last line on the form */

    /* Create field for background line */
    afield = new_field(row,0);
    ln = strlen(lp);
    if ((eline = malloc(ln+1)) == NULL)	/* allocate entry line */
	fatal(err_mem,2);
    strcpy(eline, ++lp);		/* Skip over line ID char */
    afield -> entry_line = eline;	/* Point node to it */
    
    /* In Screen mode, we'll insert into the print list as a constant datum. */

    if (screen_mode)
      { insert_in_print_list(afield);
      }
    else
      {	afield -> print_row = afield -> print_col = -1;	/* Doesn't print */
      }

    /* Look for variable data entry fields */
    for (x=0; x<ln; )		/* x is pointer to beginning of field */
    {   switch(*(lp+x))
        {   case '(':
            case '$':
            case '|':
            case '[':
                if (*(lp+x+1) != '_') {++x; break;}
            case '_':
		/* At this point, we know we're beginning a field at column x */
		afield = new_field(row,x);

		/* Set cur_field to point to the first field on this line */
		if (field_count++ == 0)
			cur_field = afield;

		/* Count, in fl, the number of characters in this field */
		start = x; fl = 1;
                while(punct(*(lp + ++x)) && x<ln ) fl++;

		memset(eline+start,' ',fl);	/* Remove from entry line */
		afield -> work = makecopy(lp+start, fl);
		afield -> mask = makecopy(lp+start, fl);
		afield -> last = makecopy(lp+start, fl);

		afield -> type = 1;	/* Assume Text */
		if (screen_mode)	/* Put all fields in print list */
			insert_in_print_list(afield);
                break;
            default:
                ++x;
        }
    }
    return(0);
}

/* LOCSCRN ------------------------------------------------

	Locate a field onto the printed page.

*/
int locscrn(lp,row,col)
char             *lp;
int row,col;
{   int ix,an_option, parens;
    char *p;

    if (row >= 0) {rw=row; cl=col;}
    pl = lp;

    if (cur_field == NULL || cur_field -> work == NULL)
	fatal_line(lp,no_field,22);	/* No fields left on this line -- error */

    if (!screen_mode && !data_mode)
      {	cur_field -> print_row = rw = newval(rw);
        cur_field -> print_col = cl = newval(cl);
	check_posn(lp);
	insert_in_print_list(cur_field);
      }

    strupr(++pl);		/* Skip over equal sign, if any */
    while (*pl != '\0')
      { an_option=keyword(field_keywords,&pl);
	switch (an_option)
	  { case 0:	/* TEXT */
	    case 1:	/* NUMERIC */
	    case 2:	/* INTEGER */
	    case 3:	/* DOLLARS */
	    case 4:	/* REAL */
	    case 5:	/* DATE */
		cur_field -> type = an_option+1;
		if (an_option == 3 && !valid_dollars(cur_field -> mask))
			fatal_line(cur_field -> mask,"~Dollar <",27);
		if (an_option == 5 && !valid_date_field(cur_field -> mask))
			fatal_line(cur_field -> mask,"~Date <",27);
		if (an_option == 2 && !valid_integer_field(cur_field->mask))
			fatal_line(cur_field -> mask,"~Integer <",27);
		break;
	    case MAX_TYPE_MODIFIER + 1:	/* UPPERCASE */
		cur_field -> ucaseonly=1;
		break;
	    case MAX_TYPE_MODIFIER + 2:	/* RIGHTJUSTIFY */
		cur_field -> rightjust=1;
		break;
	    case MAX_TYPE_MODIFIER + 3:	/* REQUIRED */
		cur_field -> reqd=(cur_field -> formula == NULL);
		/* Make it required if no formula */
		break;
	    case MAX_TYPE_MODIFIER + 4:	/* NAME */
		p = cur_field -> name;
		ix = 0;
		while (ix < 8 && *pl > ' ')
		  { *p++ = *pl++; ix++; }
		*p = '\0';
		break;
	    case MAX_TYPE_MODIFIER + 5:	/* FORMULA */
		cur_field -> prot = 1;	/* Protected field now!! */
		cur_field -> reqd = 0;	/* Can't be required. */
		while (*pl != '\0' && (isspace(*pl) || *pl == '['))
		  pl++;
		p = cur_field -> formula = malloc(strlen(pl)+1);
		parens = 0;
		while (*pl && *pl != ']')
		 { switch (*pl)
		     { case '(':
			parens++;
			break;
		       case ')':
			parens--;
			break;
		     }
		   if (!isspace(*pl))
		    *p++ = *pl++;
		   else
		    pl++;
		 }
		*p = '\0';
		while (*pl != '\0' && (isspace(*pl) || *pl == ']'))
		  pl++;
		if (parens)
		  fatal_line(cur_field->formula,"Unmatched Parentheses",26);
		break;
	    case MAX_TYPE_MODIFIER + 6:	/* NOPRINT */
		cur_field -> non_printing++;
		break;
	    case -1:	/* ??? Unknown keyword */
		fatal_line(lp,"~Keyword Specified",25);
	  }		
      }

    /* Prepare to set the next field */
    cur_field = cur_field -> next;

    return(0);
}

/* SETCNST ------------------------------------------

	Creates a constant data field.

*/


int setcnst(lp)
char             *lp;
{
struct field *constant_field;

    if (cur_field != NULL)
	didnt_posn(cur_field);

    constant_field = cur_field = new_field(-1,-1);
    if ((constant_field -> mask = strdup(lp+1)) == NULL)
	fatal(err_mem,2);
    
    return(0);
}

/* LOCCNST ------------------------------------------

	Sets the print position for a constant data field.

*/

int loccnst(lp)
char             *lp;
{
    if (cur_field == NULL || cur_field -> work != NULL)
	fatal_line(lp,no_cnst,22);	/* No nodes left to fill, or else
					   the node is a field & not a const */
    pl = lp;

    cur_field -> print_row = rw = newval(rw);
    cur_field -> print_col = cl = newval(cl);
    check_posn(lp);

    insert_in_print_list(cur_field);

    cur_field = NULL;	/* We've processed this already. */

    return(0);
}

/* LOCCTRL ------------------------------------------

	Sets the print position for a "control" constant.
	These constants occupy zero width, i.e., they print without
	moving the carriage (control & escape codes).
*/

int locctrl(lp)
char             *lp;
{
struct field *constant_field;
char *cp, constant[80];
int cntrl=0;

    if (cur_field != NULL)
	didnt_posn(cur_field);

    pl = lp;

    cur_field = NULL;
    constant_field = new_field(-1,-1);
    constant_field -> type = -1;
    constant_field -> print_row = rw = newval(rw);
    constant_field -> print_col = cl = newval(cl);
    check_posn(lp);

    cp = constant; 

    while (isspace(*pl))		/* Skip over spaces & tabs */
	pl++;
    while (*pl)				/* For each character in the string */
      {	if (cntrl)
         { cntrl=0;
	   if (*pl != CNST_CTRL)	/* Allow ~~ to quote a tilde */
		*cp++ = *pl & 31;	/* Make this character a control char */
	   else
		*cp++ = *pl;
	 }
	else
	 if (*pl == CNST_CTRL)		/* Process tilde conversion */
	   cntrl=1;
	 else
	   *cp++ = *pl;
	pl++;
      }
    *cp = '\0';

    if ((constant_field -> mask = strdup(constant)) == NULL)
	fatal(err_mem,2);

    insert_in_print_list(constant_field);

    return(0);
}

