/* restform - Part of Form Fill-r program
*/

/* #define DEBUG */

#include <math.h>
#include "globals.h"

extern go_fatal(char *fname,char *message,int err_no);
extern give_error(char *text);
extern void display_status();
extern void update_value(struct field *afield);
extern void recalc();

static void read_into_field(afield, newvalue)
  struct field *afield;
  char *newvalue;
{
  char *ps, *pw, *pm;	/* Scan across input string, work area, mask */
  int tch;		/* Length of input string */

	afield -> value = 0;	/* Initialize all value fields to zero */
	
	tch = strlen(newvalue)-1;
	if (tch >= 0 && newvalue[tch]=='\n')
		newvalue[tch]='\0';		/* Kill newline */

#ifdef DEBUG
	printf("Read_Into_Field: Print @ (%2d,%2d) <- %s\n",
		afield->print_row,afield->print_col,newvalue);
#endif

	ps = newvalue; pw = afield -> work; pm = afield -> mask;
	
	switch (afield -> type)
	 { case 1:	/* Text */
	   case 2:	/* Numeric */
	   case 6:	/* Date */
		while (*pw != '\0')
		  { if (*pm == '_')		/* Fill underlined areas */
		      if (*ps != '\0')
			*pw = *ps++;		/* ..with data, if any */
		      else
			*pw = ' ';		/* otherwise with spaces */
		    ++pm;
		    ++pw;
		  }
		/* Convert trailing spaces before underlines to underlines */
		for (tch = strlen(afield -> work), pw--, pm--;
			tch > 0 && (*pw == ' ' || *pw == '_' || *pm != '_');
			tch--, pw--, pm--)
		    if (*pm == '_') *pw = '_';
		break;
	   case 3:	/* Integer */
	   case 4:	/* Dollars */
	   case 5:	/* Real */
		afield -> value = atof(ps);
		update_value(afield);
		break;
	   default:
		break;
	 }
}

static void restore_error(a,b,c)
  char *a, *b, *c;
{ char buf[70];

	strcpy(buf,"Error");
	strcat(buf,a);
	strcat(buf,b);
	if (c != NULL)		/* 3d argument optional */
		strcat(buf,c);
	give_error(buf);
}

static char _file[]=": File ";

/* ---------------------------restform------------------------------- */
int restform()

{
FILE    *save_file;
char	save_name[81];
int     done, x;
char    tmp[132];
char    *ps;
struct field *afield;
int ret = 0;

    stat_line = stat_msg;
    strcpy(stat_msg,blanks);
    strcpy(tmp,blanks);
    if (!ext_input)
    {   do
	{   switch(showfile(save_name,
		"Select file name of form to be restored.",WFormSave,1))
	    {   case F10: return(F10);
		case 0: break;
		case 1: return(1); /* Error - no files available */
	    }
	    /* open the saved form file */
	    strupr(save_name);
	    if (strstr(save_name,FormSave) == NULL
		&& strstr(save_name,".") == NULL) strcat(save_name,FormSave);
	    if ((save_file = fopen(save_name,"r")) == NULL)
	    {   restore_error(" opening file ",save_name,NULL);
		continue;
	    }
	    if (fgets(temp1,14,save_file) == NULL)
	    {   restore_error(_file,save_name," is empty.");
		fclose(save_file);
		save_file=NULL;
		continue;
	    }
	    temp1[strlen(temp1)-1]='\0';
	    if (strstr(strupr(temp1),FormDefn) == NULL)
	    {   restore_error(_file,save_name,", first line - no .FDF filename");
		save_file=NULL;
		continue;
	    }
	    get_mode = 3;
	    x = newform();
	    get_mode=1;
	    if (x != 0)
	    {   restore_error(" recovering Form in File ",temp1,NULL);
		save_file=NULL;
		continue;
	    }
	} while (save_file == NULL);

	/* Now restore the file */

	afield = first_data_field;
	while (afield != NULL)
	  { if (afield -> work != NULL && !(afield -> prot))
	    /* Read into data fields */
	      {	if (fgets(tmp,132,save_file) == NULL) break;
		read_into_field(afield,tmp);
	      }
	    afield = afield -> next;
	  }

	recalc();

	if (fclose(save_file) != 0) ret=1;
	return(ret);
    }
    else
    {   strcpy(save_name,temp3);
	if (strstr(save_name,FormSave) == NULL
	    && strstr(save_name,".") == NULL) strcat(save_name,FormSave);
	if ((save_file = fopen(save_name,"r")) == NULL)
	    go_fatal(save_name,not_found,30);
	if (fgets(tmp,14,save_file) == NULL)
	    go_fatal(save_name," is empty",31);
	tmp[strlen(tmp)-1] = '\0';  /* Kill newline at end of string */
	ps = strstr(tmp,FormDefn);

	if (ps == NULL)		/* No .FDF in the file */
	  if (*temp1 == '\0')	/* And no .FDF on the command line */
	    go_fatal(save_name," does not contain a .FDF name",32);
	  else
	    fseek(save_file,0L,SEEK_SET); /* Use .FDF on cmd line; rewind to
					     re-read 1st line of file */
	else
	  if (*temp1 == '\0')	/* Command line overrides file .FDF */
	    strcpy(temp1,tmp);	/* Use .FDF from file if none on cmd line */

	get_mode = 3;
	if (newform()) go_fatal(temp1,not_found,30);	/* .FDF not found? */

	done = 0;
	/* Read first line to prevent extra blank form */
	if (fgets(tmp,132,save_file) == NULL) done = 1;
	while (!done)
	{
	    afield = first_data_field;
	    while (afield != NULL)
	    {   if (afield -> work != NULL && !(afield -> prot))
		/* Read into data fields */
		  { if (!done)
		      {	read_into_field(afield,tmp);
			done = (fgets(tmp,132,save_file) == NULL);
		      }
		    else
			read_into_field(afield,"");
			/* Blank fill fields not given in file */
		}
		afield = afield -> next;
	    }
	    recalc();
	    print_form();
	}
	exit(0);
    }
}
