/* Keyboard input and screen output for Z-100 computer

Routines:

int    gtkey()
void   cls()
int    getpage()
void   cursize (top_line, bot_line)short getcur()
void   scur(posn, pagenum)
short  gcur(pagenum)
void   setcur(to_row,to_col) 
short  getcur()
void   wchar(c)
void   wstr(str, move_cur)
*/

#include <stdio.h>
#include <dos.h>
#include "z100keys.h"
#include <time.h>

/* #define DEBUG 1  */

#define CTL_C        3
#define BEL          7
#define BS           8
#define CTL_Z        26
#define ESC          27

#define BDOS_IN      8  /* Used to set "Raw" mode */

int rows=24;

extern void conout(char);

/* Color attributes */
extern int    cur_attr;
extern int    nor_attr;
extern int    inv_attr;

/* gtkey ------------------------------------------------------
   Return a key from the keyboard and map control characters */

int gtkey()
{
    register int c;
    static int ateof = 0;

    if (ateof) return EOF;  /* End of input - go away */
    c=bdos(BDOS_IN,0,0) & 0xff;
    switch(c)
    {
    case 0x7F:
        return(BS);
    case 0x8D:
    case 0xCD:
        return('\n');
    case 0xE4:
        return(0x7f);
    case 0xAD:
    case 0xAE:
    case 0xB0:
    case 0xB1:
    case 0xB2:
    case 0xB3:
    case 0xB4:
    case 0xB5:
    case 0xB6:
    case 0xB7:
    case 0xB8:
    case 0xB9:
    case 0xED:
    case 0xEE:
    case 0xF0:
    case 0xF1:
    case 0xF2:
    case 0xF3:
    case 0xF4:
    case 0xF5:
    case 0xF6:
    case 0xF7:
    case 0xF8:
    case 0xF9:
        return(c & 0x3F);
    case '\r':
        return ('\n');
    case '\n':
        return ('\r');
    case CTL_C:
    case CTL_Z:
        { ateof = 1 ;; return(EOF); }
    default:
        return(c);
    }
}

/* save cursor position */
static void savecursor()
{	conout('\033');
	conout('j');
}

static void restorecursor()
{	conout('\033');
	conout('k');
} 

/* wchar -------------------------------------------------------
   Writes one character to screen */

void wchar(c)
int c;
{
    static int grafmode=0;		/* in the graphics mode? */

    static char *grafmap =
"iii`tttcct`cdddceusvabvvefusvabuusseeffbbdf\000PQqp                 g    j ^^^   ^ ";
		/* the null, P, and Q will print out as inverse
		      space, p, and q, respectively */

    c &= 255;
    switch (c)
    {
        case '\01':
	    conout('\033'); conout('m');
	    conout('0'+(inv_attr&0xF)); conout('0'+(inv_attr>>4));
	    /* conout('\033'); conout('p');*/    /* inverse video */
            break;
        case '\02':
	    conout('\033'); conout('m');
	    conout('0'+(nor_attr&0xF)); conout('0'+(nor_attr>>4));
	    /* conout('\033'); conout('q');*/    /* normal video */
            break;
        default:
	    if (c<127)
	    {	if (grafmode)
		{   conout ('\033');
		    conout ('G');
		    grafmode=0;
		}
		conout(c);
	    }
	    else
	    {   if (c>=0xB0)		/* if in the defined special chars */
		{
		    if (!grafmode)	/* put us into the graphics mode */
		    {   conout ('\033');
			conout ('F');
			grafmode++;
		    }
		    c = grafmap[c-0xB0];
		    if ((c < '^') && (c != 32))
			/* Anything but Space or Lowercase -> Inverse */
		    {	conout('\033');conout('p');
			conout(c+32);
			conout('\033');conout('q');
		    }
		    else conout(c);
		}
		else conout(32);	/* not implemented: print a space */
	    }
    }
}

/* wstr --------------------------------------------------------
   Writes a string to screen */

void wstr(str, move_cur)
char *str;
int move_cur;
{
    if (!move_cur)
	savecursor();
    while (*str) wchar(*str++);
    if (!move_cur)
	restorecursor();
}

/* cls ---------------------------------------------------------
   Clears the console screen */

void cls ()

{
	/* Note -- we have to be on the main screen, not the 25th line!! */
    wstr("\033Y8 \033K\033H\033E",1);
}

/* init_keys ---------------------------------------------------------
   Initializes the keyboard to the 8 bit mode */

void init_keys()

{ long tx;
  int got_esc=0;
    printf("\033Z");	/* Tell it to identify terminal type */

    tx = clock() + 100;		/* Wait 1/10 of a second */
    while (clock() < tx)
     if (kbhit())
      if (getch() == 27)
        got_esc++;		/* Found an ESC char, probably a Z-100 */
    if (got_esc == 0)		/* Whoa!! Must not be Z-100 */
      { fprintf(stderr,"Unknown Video Mode.");
        exit(9);
      }

/*
	Reset terminal; don't wrap lines; disable key expansion;
	block cursor; 25th line; send 'Normal Video' color sequence.
*/
    wstr("\033z\033w\033y?\033y4\033x1\002",1);

}
/* -------------------- exit_keys -----------------------------

	Restores the keyboard to the normal mode

	Removed Reset-Terminal sequence because we don't want
	to clear the screen - WL 7/26/88
*/

void exit_keys()

{
/*
	Exit graphics mode, enable key expansion, underline cursor,
	white-on-black, disable 25th line, and move to last line of display.
*/
	wstr("\033G\033x?\033y4\033m70\033y1\033Y7 ",1);
}

/* cursize -----------------------------------------------------
   sets the height of the cursor */

void cursize (kind)
int kind;
{
    if (kind) 
        wstr("\033x4",1);        /* Block cursor */
    else
        wstr("\033y4",1);        /* Underline cursor */
}

/* setcur ------------------------------------------------------
   Sets the cursor to the given position */

void setcur(to_row,to_col) 
int to_row, to_col;
{
    conout(27); conout('Y');	/* cursor addressing */
    conout(to_row+32); conout(to_col+32);
}

/* clear_eol --------------------------------------------------
   Clears to the end of the line */

void clear_eol()
{
	conout(27); conout('K');	/* Clear to end of line */
}

/*	Erase a part of the screen.  Does not move the cursor.
	If -1 is passed as min_row and max_row, the current line will
	be used.  (setcur's escape code will have an invalid line #,
	which the Z-100 will ignore.)
*/
void erase (min_row, max_row, min_col, max_col)
    int min_row, max_row, min_col, max_col;
{   int i,j;

    savecursor();
   
    for (i=min_row; i <= max_row; i++)
      {	if (max_col == 79)	/* Erase to end of line? */
	  { setcur(i,min_col);
	    clear_eol();
	  }
	else
	  if (min_col == 0)	/* Erase to start of line? */
	    { setcur(i,max_col);
	      conout(27); conout('o');
	    }
	  else			/* Erase just part of line */
	    { setcur(i,min_col);
	      for (j=min_col; j <= max_col; j++)
	        wchar(32);
	    }
      }
    restorecursor();
}

/* test main ---------------------------------------------------
   */
#if defined (DEBUG)
main()
{  int r,c;

   cls();
   setcur(5,15);
   wstr("This is a test of cls, setcur, and wstr",1);
   getchar();
   setcur(7,20);
   wstr("This is the second line of the test",0);
   getchar();
   setcur(9,25);
   wstr("This is the third line of the test",1);
   getchar();
}
#endif
